"""
Copyright 2023 [PT BOOKBOT INDONESIA](https://bookbot.id/)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os

import numpy as np
import onnxruntime

from g2p_id.onnx_utils import WrapInferenceSession

model_path = os.path.join(os.path.dirname(__file__), "models", "lstm")


class LSTM:
    """Phoneme-level LSTM model for sequence-to-sequence phonemization.
    Trained with [Keras](https://keras.io/examples/nlp/lstm_seq2seq/),
    and exported to ONNX. ONNX Runtime engine used during inference.
    """

    def __init__(self):
        encoder_model_path = os.path.join(model_path, "encoder_model.onnx")
        decoder_model_path = os.path.join(model_path, "decoder_model.onnx")
        g2id_path = os.path.join(model_path, "g2id.json")
        p2id_path = os.path.join(model_path, "p2id.json")
        config_path = os.path.join(model_path, "config.json")
        self.encoder = WrapInferenceSession(
            encoder_model_path,
            providers=onnxruntime.get_available_providers(),
        )
        self.decoder = WrapInferenceSession(
            decoder_model_path,
            providers=onnxruntime.get_available_providers(),
        )
        with open(g2id_path, encoding="utf-8") as file:
            self.g2id = json.load(file)
        with open(p2id_path, encoding="utf-8") as file:
            self.p2id = json.load(file)
        self.id2p = {v: k for k, v in self.p2id.items()}
        with open(config_path, encoding="utf-8") as file:
            self.config = json.load(file)

    def predict(self, text: str) -> str:
        """Performs LSTM inference, predicting phonemes of a given word.

        Args:
            text (str): Word to convert to phonemes.

        Returns:
            str: Word in phonemes.
        """
        input_seq = np.zeros(
            (
                1,
                self.config["max_encoder_seq_length"],
                self.config["num_encoder_tokens"],
            ),
            dtype="float32",
        )

        for idx, char in enumerate(text):
            input_seq[0, idx, self.g2id[char]] = 1.0
        input_seq[0, len(text) :, self.g2id[self.config["pad_token"]]] = 1.0

        encoder_inputs = {"input_1": input_seq}
        states_value = self.encoder.run(None, encoder_inputs)

        target_seq = np.zeros((1, 1, self.config["num_decoder_tokens"]), dtype="float32")
        target_seq[0, 0, self.p2id[self.config["bos_token"]]] = 1.0

        stop_condition = False
        decoded_sentence = ""
        while not stop_condition:
            decoder_inputs = {
                "input_2": target_seq,
                "input_3": states_value[0],
                "input_4": states_value[1],
            }
            output_tokens, state_memory, state_carry = self.decoder.run(None, decoder_inputs)

            sampled_token_index = np.argmax(output_tokens[0, -1, :])
            sampled_char = self.id2p[sampled_token_index]
            decoded_sentence += sampled_char

            if (
                sampled_char == self.config["eos_token"]
                or len(decoded_sentence) > self.config["max_decoder_seq_length"]
            ):
                stop_condition = True

            target_seq = np.zeros((1, 1, self.config["num_decoder_tokens"]), dtype="float32")
            target_seq[0, 0, sampled_token_index] = 1.0

            states_value = [state_memory, state_carry]

        return decoded_sentence.replace(self.config["eos_token"], "")



"""
Copyright 2023 [PT BOOKBOT INDONESIA](https://bookbot.id/)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import pickle
import unicodedata
from builtins import str as unicode
from itertools import permutations
from typing import Dict, List, Tuple, Union

import nltk
from nltk.tag.perceptron import PerceptronTagger
from nltk.tokenize import TweetTokenizer

from g2p_id.bert import BERT
from g2p_id.lstm import LSTM
from g2p_id.text_processor import TextProcessor

nltk.download("wordnet")
resources_path = os.path.join(os.path.dirname(__file__), "resources")


def construct_homographs_dictionary() -> Dict[str, Tuple[str, str, str, str]]:
    """Creates a dictionary of homographs

    Returns:
        Dict[str, Tuple[str, str, str, str]]:
            Key: WORD
            Value: (PH1, PH2, POS1, POS2)
    """
    homograph_path = os.path.join(resources_path, "homographs_id.tsv")
    homograph2features = {}
    with open(homograph_path, encoding="utf-8") as file:
        lines = file.readlines()
        for line in lines:
            grapheme, phone_1, phone_2, pos_1, pos_2 = line.strip("\n").split("\t")
            homograph2features[grapheme.lower()] = (phone_1, phone_2, pos_1, pos_2)

    return homograph2features


def construct_lexicon_dictionary() -> Dict[str, str]:
    """Creates a lexicon dictionary.

    Returns:
        Dict[str, str]:
            Key: WORD
            Value: Phoneme (IPA)
    """
    lexicon_path = os.path.join(resources_path, "lexicon_id.tsv")
    lexicon2features = {}
    with open(lexicon_path, encoding="utf-8") as file:
        lines = file.readlines()
        for line in lines:
            grapheme, phoneme = line.strip("\n").split("\t")
            lexicon2features[grapheme.lower()] = phoneme
    return lexicon2features


class G2p:
    """Grapheme-to-phoneme (g2p) main class for phonemization.
    This class provides a high-level API for grapheme-to-phoneme conversion.

    1. Preprocess and normalize text
    2. Word tokenizes text
    3. Predict POS for every word
    4. If word is non-alphabetic, add to list (i.e. punctuation)
    5. If word is a homograph, check POS and use matching word's phonemes
    6. If word is a non-homograph, lookup lexicon
    7. Otherwise, predict with a neural network
    """

    def __init__(self, model_type="BERT"):
        """Constructor for G2p.

        Args:
            model_type (str, optional):
                Type of neural network to use for prediction.
                Choices are "LSTM" or "BERT". Defaults to "BERT".
        """
        self.homograph2features = construct_homographs_dictionary()
        self.lexicon2features = construct_lexicon_dictionary()
        self.normalizer = TextProcessor()
        self.tagger = PerceptronTagger(load=False)
        tagger_path = os.path.join(resources_path, "id_posp_tagger.pickle")
        with open(tagger_path, "rb") as f:
            self.tagger = self.tagger.decode_json_obj(pickle.load(f))
        self.model: Union[BERT, LSTM] = BERT() if model_type == "BERT" else LSTM()
        self.tokenizer = TweetTokenizer()
        self.pos_dict = {
            "N": ["B-NNO", "B-NNP", "B-PRN", "B-PRN", "B-PRK"],
            "V": ["B-VBI", "B-VBT", "B-VBP", "B-VBL", "B-VBE"],
            "A": ["B-ADJ"],
            "P": ["B-PAR"],
        }

    def _preprocess(self, text: str) -> str:
        """Performs preprocessing.
        (1) Adds spaces in between tokens
        (2) Normalizes unicode and accents
        (3) Normalizes numbers
        (4) Lower case texts
        (5) Removes unwanted tokens

        Arguments:
            text (str): Text to preprocess.

        Returns:
            str: Preprocessed text.
        """
        text = text.replace("-", " ")
        text = re.sub(r"\.(?=.*\.)", " ", text)
        text = " ".join(self.tokenizer.tokenize(text))
        text = unicode(text)
        text = "".join(char for char in unicodedata.normalize("NFD", text) if unicodedata.category(char) != "Mn")
        text = self.normalizer.normalize(text).strip()
        text = text.lower()
        text = re.sub(r"[^ a-z'.,?!\-]", "", text)
        return text

    def _rule_based_g2p(self, text: str) -> str:
        """Applies rule-based Indonesian grapheme2phoneme conversion.

        Args:
            text (str): Grapheme text to convert to phoneme.

        Returns:
            str: Phoneme string.
        """
        phonetic_mapping = {
            "ny": "ɲ",
            "ng": "ŋ",
            "sy": "ʃ",
            "aa": "aʔa",
            "ii": "iʔi",
            "oo": "oʔo",
            "əə": "əʔə",
            "uu": "uʔu",
            "'": "ʔ",
            "g": "ɡ",
            "q": "k",
            "j": "dʒ",
            "y": "j",
            "x": "ks",
            "c": "tʃ",
            "kh": "x",
        }

        if text.startswith("x"):
            text = "s" + text[1:]

        if text.startswith("ps"):
            text = text[1:]

        for graph, phone in phonetic_mapping.items():
            text = text.replace(graph, phone)

        phonemes = [list(phn) if phn not in ("dʒ", "tʃ") else [phn] for phn in re.split("(tʃ|dʒ)", text)]
        return " ".join([p for phn in phonemes for p in phn])

    def __call__(self, text: str) -> List[List[str]]:
        """Grapheme-to-phoneme converter.

        1. Preprocess and normalize text
        2. Word tokenizes text
        3. Predict POS for every word
        4. If word is non-alphabetic, add to list (i.e. punctuation)
        5. If word is a homograph, check POS and use matching word's phonemes
        6. If word is a non-homograph, lookup lexicon
        7. Otherwise, predict with a neural network

        Args:
            text (str): Grapheme text to convert to phoneme.

        Returns:
            List[List[str]]: List of strings in phonemes.
        """
        text = self._preprocess(text)
        words = self.tokenizer.tokenize(text)
        tokens = self.tagger.tag(words)

        prons = []
        for word, pos in tokens:
            pron = ""
            if re.search("[a-z]", word) is None:  # non-alphabetic
                pron = word

            elif word in self.homograph2features:  # check if homograph
                pron1, pron2, pos1, _ = self.homograph2features[word]

                # check for the matching POS
                if pos in self.pos_dict[pos1]:
                    pron = pron1
                else:
                    pron = pron2

            elif word in self.lexicon2features:  # non-homographs
                pron = self.lexicon2features[word]

            else:  # predict for OOV
                pron = self.model.predict(word)
                if isinstance(self.model, BERT):
                    pron = self._rule_based_g2p(pron)

            if pron.endswith("ʔ"):
                pron = pron[:-1] + "k"

            consonants = "bdjklmnprstwɲ"
            vowels = "aeiouə"

            for letter in consonants:
                pron = pron.replace(f"ʔ {letter}", f"k {letter}")

            # add a glottal stop in between consecutive vowels
            for v1, v2 in permutations(vowels, 2):
                pron = pron.replace(f"{v1} {v2}", f"{v1} ʔ {v2}")

            prons.append(pron.split())

        return prons



"""
Copyright 2023 [PT BOOKBOT INDONESIA](https://bookbot.id/)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os

import numpy as np
import onnxruntime

from g2p_id.onnx_utils import WrapInferenceSession

model_path = os.path.join(os.path.dirname(__file__), "models", "bert")


class BERT:
    """Phoneme-level BERT model for predicting the correct phoneme for the letter `e`.
    Trained with [Keras](https://keras.io/examples/nlp/masked_language_modeling/),
    and exported to ONNX. ONNX Runtime engine used during inference.
    """

    def __init__(self):
        bert_model_path = os.path.join(model_path, "bert_mlm.onnx")
        token2id = os.path.join(model_path, "token2id.json")
        config_path = os.path.join(model_path, "config.json")
        self.model = WrapInferenceSession(bert_model_path, providers=onnxruntime.get_available_providers())
        with open(config_path, encoding="utf-8") as file:
            self.config = json.load(file)
        with open(token2id, encoding="utf-8") as file:
            self.token2id = json.load(file)
        self.id2token = {v: k for k, v in self.token2id.items()}

    def predict(self, text: str) -> str:
        """Performs BERT inference, predicting the correct phoneme for the letter `e`.

        Args:
            text (str): Word to predict from.

        Returns:
            str: Word after prediction.
        """
        # `x` is currently OOV, we replace with
        text = text.replace("x", "ks")
        # mask `e`'s
        text = " ".join([c if c != "e" else "[mask]" for c in text])

        # tokenize and pad to max length
        tokens = [self.token2id[c] for c in text.split()]
        padding = [self.token2id[self.config["pad_token"]] for _ in range(self.config["max_seq_length"] - len(tokens))]
        tokens = tokens + padding

        input_ids = np.array([tokens], dtype="int64")
        inputs = {"input_1": input_ids}
        prediction = self.model.run(None, inputs)

        # find masked idx token
        mask_token_id = self.token2id[self.config["mask_token"]]
        masked_index = np.where(input_ids == mask_token_id)[1]

        # get prediction at masked indices
        mask_prediction = prediction[0][0][masked_index]
        predicted_ids = np.argmax(mask_prediction, axis=1)

        # replace mask with predicted token
        for i, idx in enumerate(masked_index):
            tokens[idx] = predicted_ids[i]

        return "".join([self.id2token[t] for t in tokens if t != 0])



def test_g2p(g2p):
    assert g2p("Apel itu berwarna merah.") == [
        ["a", "p", "ə", "l"],
        ["i", "t", "u"],
        ["b", "ə", "r", "w", "a", "r", "n", "a"],
        ["m", "e", "r", "a", "h"],
        ["."],
    ]
    assert g2p("Rahel bersekolah di S M A Jakarta 17.") == [
        ["r", "a", "h", "e", "l"],
        ["b", "ə", "r", "s", "ə", "k", "o", "l", "a", "h"],
        ["d", "i"],
        ["e", "s"],
        ["e", "m"],
        ["a"],
        ["dʒ", "a", "k", "a", "r", "t", "a"],
        ["t", "u", "dʒ", "u", "h"],
        ["b", "ə", "l", "a", "s"],
        ["."],
    ]
    assert g2p("Mereka sedang bermain bola di lapangan.") == [
        ["m", "ə", "r", "e", "k", "a"],
        ["s", "ə", "d", "a", "ŋ"],
        ["b", "ə", "r", "m", "a", "ʔ", "i", "n"],
        ["b", "o", "l", "a"],
        ["d", "i"],
        ["l", "a", "p", "a", "ŋ", "a", "n"],
        ["."],
    ]
    assert g2p("Ini rumahnya Aisyah dan Ceri.") == [
        ["i", "n", "i"],
        ["r", "u", "m", "a", "h", "ɲ", "a"],
        ["a", "ʔ", "i", "ʃ", "a", "h"],
        ["d", "a", "n"],
        ["tʃ", "e", "r", "i"],
        ["."],
    ]
    assert g2p("keset selamat datang") == [
        ["k", "e", "s", "e", "t"],
        ["s", "ə", "l", "a", "m", "a", "t"],
        ["d", "a", "t", "a", "ŋ"],
    ]
    assert g2p("kakak layak") == [["k", "a", "k", "a", "k"], ["l", "a", "j", "a", "k"]]


def test_rule_based_g2p(g2p):
    assert g2p._rule_based_g2p("berakhirnya") == "b e r a x i r ɲ a"
    assert g2p._rule_based_g2p("bermaaf-maafan") == "b e r m a ʔ a f - m a ʔ a f a n"
    assert g2p._rule_based_g2p("kecolongan") == "k e tʃ o l o ŋ a n"
    assert g2p._rule_based_g2p("jayapura") == "dʒ a j a p u r a"
    assert g2p._rule_based_g2p("xenon") == "s e n o n"
    assert g2p._rule_based_g2p("layak") == "l a j a k"


def test_lstm(lstm):
    assert lstm.predict("mengembangkannya") == "məŋəmbaŋkanɲa"
    assert lstm.predict("merdeka") == "mərdeka"
    assert lstm.predict("pecel") == "pətʃəl"
    assert lstm.predict("lele") == "lele"


def test_bert(bert):
    assert bert.predict("mengembangkannya") == "məngəmbangkannya"
    assert bert.predict("merdeka") == "mərdeka"
    assert bert.predict("pecel") == "pəcel"
    assert bert.predict("lele") == "lele"
    assert bert.predict("banyak") == "banyak"


def test_ps(g2p):
    assert g2p("psikologi") == [["s", "i", "k", "o", "l", "o", "ɡ", "i"]]
    assert g2p("psikometri") == [["s", "i", "k", "o", "m", "e", "t", "r", "i"]]
    assert g2p("psikotes") == [["s", "i", "k", "o", "t", "e", "s"]]


def test_sticking_dot(g2p):
    assert g2p("Seniornya Brigadir Jendral A.Yani mengambil alih pimpinan.") == [
        ["s", "ə", "n", "i", "ʔ", "o", "r", "ɲ", "a"],
        ["b", "r", "i", "ɡ", "a", "d", "i", "r"],
        ["dʒ", "ə", "n", "d", "r", "a", "l"],
        ["a"],
        ["j", "a", "n", "i"],
        ["m", "ə", "ŋ", "a", "m", "b", "i", "l"],
        ["a", "l", "i", "h"],
        ["p", "i", "m", "p", "i", "n", "a", "n"],
        ["."],
    ]


def test_onnx_wrapper(bert):
    assert bert.predict("mengembangkannya") == "məngəmbangkannya"
    model_state = bert.model.__getstate__()
    bert.model.__setstate__(model_state)
    assert bert.predict("mengembangkannya") == "məngəmbangkannya"



def test_text_processor(text_processor):
    # URLs
    assert text_processor.normalize("Situs: https://www.google.com") == "Situs: "
    # measurements
    assert (
        text_processor.normalize("123,1 kg")
        == "seratus dua puluh tiga koma satu kilogram"
    )
    assert text_processor.normalize("500 cm") == "lima ratus centimeter"
    # currency/money
    assert text_processor.normalize("$100") == "seratus dollar"
    assert text_processor.normalize("Rp 3,000,000") == "tiga juta rupiah"
    # dates
    assert (
        text_processor.normalize("(17/8/1945)").strip()
        == "tujuh belas Agustus seribu sembilan ratus empat puluh lima"
    )
    assert text_processor.normalize("(1/13)").strip() == "satu Januari"
    # time/time zone
    assert (
        text_processor.normalize("19.45 WIB")
        == "sembilan belas lewat empat puluh lima menit Waktu Indonesia Barat"
    )
    assert (
        text_processor.normalize("19.00 WIB") == "sembilan belas Waktu Indonesia Barat"
    )
    # numerics
    assert text_processor.normalize("105.000") == "seratus lima ribu"
    assert text_processor.normalize("0,5") == "nol koma lima"



from setuptools import find_packages, setup
from pathlib import Path

this_path = Path(__file__).parent

readme_path = this_path / "README.md"
requirements_path = this_path / "requirements.txt"

long_description = readme_path.read_text(encoding="utf-8")

with open(requirements_path, "r", encoding="utf-8") as requirements_file:
    requirements = requirements_file.read().splitlines()

if __name__ == "__main__":
    setup(
        name="g2p_id_py",
        version="0.4.2",
        description="Indonesian G2P.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="w11wo",
        author_email="wilson@bookbotkids.com",
        url="https://github.com/bookbot-kids/g2p_id",
        license="Apache License",
        packages=find_packages(),
        install_requires=requirements,
        include_package_data=True,
        platforms=["linux", "unix", "windows"],
        python_requires=">=3.8",
    )



"""
MIT License

Copyright (c) 2021 Cahya Wirawan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import os
import re
from typing import Any

from num2words import num2words

resources_path = os.path.join(os.path.dirname(__file__), "resources")


class TextProcessor:
    """Indonesian text processor to normalize numerics, currencies, and timezones."""

    def __init__(self):
        self.measurements = {}
        self.thousands = ["ratus", "ribu", "juta", "miliar", "milyar", "triliun"]
        self.months = [
            "Januari",
            "Februari",
            "Maret",
            "April",
            "Mei",
            "Juni",
            "Juli",
            "Agustus",
            "September",
            "Oktober",
            "November",
            "Desember",
        ]
        measurements_path = os.path.join(resources_path, "measurements.tsv")
        currencies_path = os.path.join(resources_path, "currency.tsv")
        timezones_path = os.path.join(resources_path, "timezones.tsv")

        with open(measurements_path, "r", encoding="utf-8") as file:
            for lines in file:
                line = lines.strip().split("\t")
                self.measurements[line[0]] = line[1]

        self.currencies = {}
        with open(currencies_path, "r", encoding="utf-8") as file:
            for lines in file:
                line = lines.strip().split("\t")
                self.currencies[line[0]] = line[1]

        self.timezones = {}
        with open(timezones_path, "r", encoding="utf-8") as file:
            for lines in file:
                line = lines.strip().split("\t")
                self.timezones[line[0]] = line[1]

        self.re_thousands = "|".join(self.thousands)
        self.re_currencies = r"\b" + re.sub(
            r"\|([^|$£€¥₩]+)", r"|\\b\1", "|".join(list(self.currencies))
        )
        self.re_currencies = re.sub(r"([$£€¥₩])", r"\\\1", self.re_currencies)
        self.re_moneys = (
            rf"(({self.re_currencies}) ?([\d\.\,]+)( ({self.re_thousands})?(an)?)?)"
        )
        self.re_measurements = "|".join(list(self.measurements))
        self.re_measurements = rf"(\b([\d\.\,]+) ?({self.re_measurements})\b)"
        self.re_timezones = "|".join(list(self.timezones))
        self.re_timezones = (
            r"((\d{1,2})[\.:](\d{1,2}) " + rf"\b({self.re_timezones})\b)"
        )
        self.re_http = re.compile(
            r"""
            (https?://(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.
            [a-zA-Z0-9()]{1,6}\b[-a-zA-Z0-9()@:%_\+.~#?&//=]*)
            """,
            re.X,
        )

    @staticmethod
    def is_integer(number: Any) -> bool:
        """Check if integer by type-casting.

        Args:
            number (Any): Number to check.

        Returns:
            bool: Is a valid integer.
        """
        try:
            int(number)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_float(number: Any) -> bool:
        """Check if float by type-casting.

        Args:
            number (Any): Number to check.

        Returns:
            bool: Is a valid float.
        """
        try:
            float(number)
            return True
        except ValueError:
            return False

    def normalize_url(self, text: str) -> str:
        """Removes URL from text.

        Args:
            text (str): Text with URL to normalize.

        Returns:
            str: Normalized text with URLs removed.
        """
        urls = re.findall(self.re_http, text)
        for url in urls:
            text = text.replace(url[0], "")
        return text

    def normalize_currency(self, text: str) -> str:
        """Normalizes international and Indonesian (Rupiah) currencies.

        Examples:
        - `"$250"` -> `"dua ratus lima puluh dollar"`
        - `"Rp 3,000,000"` -> `"tiga juta rupiah"`

        Args:
            text (str): Text with currency to normalize.

        Returns:
            str: Normalized text with currency transliterated.
        """
        moneys = re.findall(self.re_moneys, text)
        for money in moneys:
            number: Any = re.sub(",", ".", re.sub(r"\.", "", money[2].strip(" ,.")))
            try:
                if number == "":
                    continue
                if self.is_integer(number):
                    number = int(number)
                elif self.is_float(number):
                    number = float(number)
                else:
                    number = re.sub(r"[.,]", "", number)
                    number = int(number)
                number = num2words(number, to="cardinal", lang="id")
                text = text.replace(
                    money[0].strip(" ,."),
                    f"{number} {money[3]} {self.currencies[money[1]]}",
                )
            except NotImplementedError as error:
                print(error)
                print(f"Problem with money: <{text}>: {number}")
        return text

    def normalize_measurement(self, text: str) -> str:
        """Normalizes measurement units, including its scalar value.

        Examples:
        - `"10,5 km"` -> `"sepuluh koma lima kilometer"`
        - `"5°C"` -> `"lima derajat celsius"`

        Args:
            text (str): Text with measurements to normalize.

        Returns:
            str: Normalized text with measurements transliterated.
        """
        units = re.findall(self.re_measurements, text)
        for unit in units:
            number: Any = re.sub(",", ".", re.sub(r"\.", "", unit[1].strip(" ,.")))
            try:
                if number == "":
                    continue
                if re.search(r"\.", number):
                    number = float(number)
                else:
                    number = int(number)
                number = num2words(number, to="cardinal", lang="id")
                text = text.replace(
                    unit[0].strip(" ,."), f"{number} {self.measurements[unit[2]]}"
                )
            except NotImplementedError as error:
                print(error)
                print(f"Problem with measurements: <{text}>: {number}")
        return text

    def normalize_date(self, text: str) -> str:
        """Normalizes dates.

        Examples:
        - `"(12/3/2021)"` -> `"dua belas Maret dua ribu dua puluh satu"`

        Args:
            text (str): Text with dates to normalize.

        Returns:
            str: Normalized text with dates transliterated.
        """
        dates = re.findall(r"(\((\d{1,2})/(\d{1,2})(/(\d+))?\))", text)
        for date in dates:
            try:
                day = num2words(int(date[1]), to="cardinal", lang="id")
                month: Any = int(date[2]) - 1
                if month >= 12:
                    month = 0
                month = self.months[month]
                if date[4] != "":
                    year = num2words(int(date[4]), to="cardinal", lang="id")
                    date_string = f"{day} {month} {year}"
                else:
                    date_string = f"{day} {month}"
                text = text.replace(date[0], f" {date_string} ")
            except NotImplementedError as error:
                print(error)
                print(f"Problem with dates: <{text}>: {date}")
        return text

    def normalize_timezone(self, text: str) -> str:
        """Normalizes Indonesian time with timezones.

        Examples:
        - `"22.30 WITA"`
            -> `"dua puluh dua lewat tiga puluh menit Waktu Indonesia Tengah"`

        Args:
            text (str): Text with timezones to normalize.

        Returns:
            str: Normalized text with timezones transliterated.
        """
        timezones = re.findall(self.re_timezones, text)
        for timezone in timezones:
            try:
                hour = num2words(int(timezone[1]), to="cardinal", lang="id")
                minute = num2words(int(timezone[2]), to="cardinal", lang="id")
                zone = self.timezones[timezone[3]]
                if minute == "nol":
                    time_string = f"{hour} {zone}"
                else:
                    time_string = f"{hour} lewat {minute} menit {zone}"
                text = text.replace(timezone[0], f"{time_string}")
            except NotImplementedError as error:
                print(error)
                print(f"Problem with timezones: <{text}>: {timezone}")
        return text

    def normalize_number(self, text: str) -> str:
        """Normalizes Arabic numbers to Indonesian.

        Examples:
        - `"1.000"` -> `"seribu"`
        - `"10,5"` -> `"sepuluh koma lima"`

        Args:
            text (str): Text with numbers to normalize.

        Returns:
            str: Normalized text with numbers transliterated.
        """
        re_numbers = [r"([\d.,]+)", r"\d+"]
        for re_number in re_numbers:
            number_len = 0
            for i in re.finditer(re_number, text):
                start = i.start() + number_len
                end = i.end() + number_len
                number: Any = text[start:end]
                number = re.sub(",", ".", re.sub(r"\.", "", number.strip(" ,.")))
                if number == "":
                    continue
                if self.is_float(number) or self.is_integer(number):
                    try:
                        if self.is_integer(number):
                            number = int(number)
                        else:
                            number = float(number)
                        number = num2words(number, to="cardinal", lang="id")
                        text = text[:start] + number + text[end:]
                        number_len += len(number) - (end - start)
                    except NotImplementedError as error:
                        print(error)
                        print(f"Problem with number: <{text}>: {number}")
        return text

    def normalize(self, text: str) -> str:
        """Normalizes Indonesian text by expanding:

        - URL
        - Currency
        - Measurements
        - Dates
        - Timezones
        - Arabic Numerals

        Args:
            text (str): Text to normalize.

        Returns:
            str: Normalized text.
        """
        # Remove URL
        text = self.normalize_url(text)
        # Currency
        text = self.normalize_currency(text)
        # Measurements
        text = self.normalize_measurement(text)
        # Date
        text = self.normalize_date(text)
        # Timezones
        text = self.normalize_timezone(text)
        # Any number
        text = self.normalize_number(text)
        # collapse consecutive whitespaces
        text = re.sub(r"\s+", " ", text)
        return text



"""
Copyright 2023 [PT BOOKBOT INDONESIA](https://bookbot.id/)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import onnxruntime as ort


class WrapInferenceSession:
    """Wrapper class for serializing ONNX InferenceSession objects.
    Based on: https://github.com/microsoft/onnxruntime/pull/800#issuecomment-844326099
    """

    def __init__(self, onnx_bytes, sess_options=None, providers=None):
        self.sess = ort.InferenceSession(onnx_bytes, sess_options=sess_options, providers=providers)
        self.onnx_bytes = onnx_bytes
        self.providers = providers

    def run(self, *args):
        """Wrapper for ONNX InferenceSession run method.

        Returns:
            Any: Inference result.
        """
        return self.sess.run(*args)

    def __getstate__(self):
        return {"onnx_bytes": self.onnx_bytes}

    def __setstate__(self, values):
        self.onnx_bytes = values["onnx_bytes"]
        self.providers = values.get("providers", None)
        self.sess = ort.InferenceSession(self.onnx_bytes, self.providers)
